# ===================================================================================================================== #
# SAFIM: Semi-Analytical Floating Iceberg Model										#
# by Vliadislav A. Yastrebov(1) and Pauline Bonnet(1,2,3)								#
# (1) MINES ParisTech, PSL University, Centre des materiaux, CNRS UMR 7633, Evry, France				#
# (2) Laboratoire Procedes et Ingenierie en Mecanique et Materiaux, CNRS, ENSAM, CNAM, Paris, France                    #
# (3) Institut de Physique du Globe de Paris, Sorbonne Paris Cite, Seismology Group, CNRS UMR 7154, Paris, France    	#
# v0.0.0 29/05/2017													#
# v1.0.0 04/12/2019													#
# v1.0.1 18/12/2019 (improved indentation)										#
# (cc) Creative Commons Attribution 4.0 International									#
# ===================================================================================================================== #

import numpy as np
#import Functions_iceberg_simu_alleps as ff
import os

# Iceberg class
class Iceberg:
# Constructor parameters:
# 	initial coordinates: r0_ = [x0,y0,0] (m)
# 	initial tilt: theta0 = [0,0,th] (rad) 
# 	width of the iceberg : L_ (m)
# 	height of the iceberg : H (m)
# 	damping scaling : alpha_ (-)
# 	ice density : rho_i_ (kg/m^3)
# 	water density : rho_w_ (kg/m^3)
    def __init__(self,r0_,theta0,L_,H_,alpha_,rho_i_,rho_w_): 
        self.r0 = r0_
        self.theta0 = theta0
        self.theta = theta0
        self.H = H_
        self.L = L_
        self.segments = [0,1,2,3]
        self.lseg  = [self.L,self.H,self.L,self.H]
        self.rho_i = rho_i_
        self.rho_w = rho_w_
	# Linear mass density
        self.mass = self.rho_i * self.L * self.H
	# Moment of inertia around the center of mass
        self.inert = self.mass * (self.L**2 + self.H**2) / 12.
        self.alpha = alpha_
	# Water level
        self.z0 = 0.
	# Acceleration of gravity (m/s)
        self.g = 9.81
	# Contact, define the wall position
        self.ex=np.array([1,0,0])
        self.ez=np.array([0,1,0])
        r0=np.array([0.,0.,0.]) # center of mass 
        theta=np.array([0.,0.,0.]) # rotation angle, zero for vertical iceberg  
	# Coordinates of corners
        r1,r2,r3,r4 = self.getRect()
        lx = np.array([r1[0],r2[0],r3[0],r4[0]])
	# The most left corner
        self.x_wall = np.sort(lx)[0]

    # Return normal vector for a given tangent vector
    def norm(self,tang):
        n = tang.copy()
        n[0] = tang[1]
        n[1] = -tang[0]
        n[2] = 0.
        return n/np.sqrt(np.dot(n,n))

    # Sign function
    def sign(self,x):
        if x>0:
            return 1.
        elif x<0:
            return -1.
        else:
            return 1.

    # Added moment of inertia around the center of mass.
    # From: translation in English of PhD thesis of Kurt Wendel "Hydrodynamic masses and hydrodynamic moments of inertia" (1956)
    # available on https://dome.mit.edu/handle/1721.3/51294
    def Hinert(self):
        return 0.1335 * np.pi * self.rho_w * (self.H/2.)**4 # average between 0.117 and 0.150

    # Added mass in OX direction. The same source.
    def Hmx(self):
        r1,r2,r3,r4 = self.getRect()
        hy=np.array([r1[1],r2[1],r3[1],r4[1]])
        hys=np.sort(hy)
        heff = (self.z0-hys[0])
        return 0.25*np.pi*self.rho_w*heff**2

    # Added mass in OY (or OZ) direction. The same source.
    def Hmz(self):  
        r1,r2,r3,r4 = self.getRect()
        hx=np.array([r1[0],r2[0],r3[0],r4[0]])
        under_water = np.array([])
        hy=np.array([r1[1],r2[1],r3[1],r4[1]])
        for i in range(1,hy.shape[0]):            
            if hy[i] < self.z0:
                under_water = np.append(under_water,hx[i])
            xi = (self.z0-hy[i-1]) / (hy[i]-hy[i-1])
            if xi >= 0 and xi <= 1:
                under_water = np.append(under_water,hx[i-1]+xi*(hx[i]-hx[i-1]))

        if hy[0] < self.z0:
            under_water = np.append(under_water,hx[0])
        xi = (self.z0-hy[3]) / (hy[0]-hy[3])
        if xi >= 0 and xi <= 1:
            under_water = np.append(under_water,hx[3]+xi*(hx[0]-hx[3]))
        under_water = np.sort(under_water)
        return 0.75*np.pi*self.rho_w*( (under_water[-1] - under_water[0])*0.5 )**2

    # Returns coordinates of corners
    def getRect(self):
        e1=self.ex*np.cos(self.theta[2])+self.ez*np.sin(self.theta[2])
        e2=-self.ex*np.sin(self.theta[2])+self.ez*np.cos(self.theta[2])
        v1 = 0.5*(-e1*self.L-e2*self.H)
        v2 = 0.5*(e1*self.L-e2*self.H)
        v3 = 0.5*(e1*self.L+e2*self.H)
        v4 = 0.5*(-e1*self.L+e2*self.H)
        return self.r0+v1,self.r0+v2,self.r0+v3,self.r0+v4
   
    # Returns normal vector and corners for the selected segment 0,1,2 or 3. 
    def seg(self,i):
        e1=self.ex*np.cos(self.theta[2])+self.ez*np.sin(self.theta[2])
        e2=-self.ex*np.sin(self.theta[2])+self.ez*np.cos(self.theta[2])
        v1 = 0.5*(-e1*self.L-e2*self.H)
        v2 = 0.5*(e1*self.L-e2*self.H)
        v3 = 0.5*(e1*self.L+e2*self.H)
        v4 = 0.5*(-e1*self.L+e2*self.H)
        if i == 0:
            tang = v2-v1
            return self.norm(tang),self.r0+v1,self.r0+v2
        elif i == 1:
            return self.norm(v3-v2),self.r0+v2,self.r0+v3
        elif i == 2:
            return self.norm(v4-v3),self.r0+v3,self.r0+v4
        elif i == 3:
            return self.norm(v1-v4),self.r0+v4,self.r0+v1
        else:
            print ('Error')
            return np.array([0.,0.,0.]),np.array([0.,0.,0.]),np.array([0.,0.,0.])
        
    # Returns a force vector acting on the iceberg from gravity and hydrostatic pressure and contact
    def F(self):
        F_=np.array([0.,0.,0.])
        Fcontact=0
        for s in self.segments:
                Fc=np.array([0.,0.,0.])                
                n,r1,r2=self.seg(s)
                z1=np.dot(r1,self.ez)
                z2=np.dot(r2,self.ez) 
                if z1<=self.z0 and z2<=self.z0:
                    Fc=n*self.rho_w*self.g*self.lseg[s]*(0.5*(z1+z2) - self.z0)
                elif z1<=self.z0 and z2>self.z0:
                    xi=(self.z0-z1)/(z2-z1)
                    Fc=n*self.rho_w*self.g*self.lseg[s]*((z1-self.z0)*xi+(z2-z1)*xi**2/2.)
                elif z1>self.z0 and z2<=self.z0:
                    xi=(self.z0-z1)/(z2-z1)
                    Fc=n*self.rho_w*self.g*self.lseg[s]*((z1-self.z0)*(1-xi)+(z2-z1)*(1/2.-xi**2/2.))
                else:
                    Fc=np.array([0.,0.,0.])
                F_ += Fc
        return F_ - self.mass*self.g*self.ez

    # Returns a moment acting on the iceberg from hydrostatic pressure and computed around its center of mass                    
    def M(self):
        M_=np.array([0.,0.,0.])
        Mcontact=np.array([0.,0.,0.])
        for s in self.segments:
                Mc=np.array([0.,0.,0.])
                n,r1,r2=self.seg(s)
                z1=np.dot(r1,self.ez)            
                z2=np.dot(r2,self.ez) 
                if z1<=self.z0 and z2<=self.z0:
                    Integ = (z1-self.z0)*(r1-self.r0)*1+((z1-self.z0)*(r2-r1)+(z2-z1)*(r1-self.r0))*(1**2/2.-0.**2/2.)+(z2-z1)*(r2-r1)*(1**3/3.-0.**3/3.)
                    Mc=self.rho_w*self.g*self.lseg[s]*np.cross(Integ,n)
                elif z1<=self.z0 and z2>self.z0:
                    xi=(self.z0-z1)/(z2-z1)
                    Integ = (z1-self.z0)*(r1-self.r0)*xi+((z1-self.z0)*(r2-r1)+(z2-z1)*(r1-self.r0))*(xi**2/2.-0.**2/2.)+(z2-z1)*(r2-r1)*(xi**3/3.-0.**3/3.)
                    Mc=self.rho_w*self.g*self.lseg[s]*np.cross(Integ,n)
                elif z1>self.z0 and z2<=self.z0:
                    xi=(self.z0-z1)/(z2-z1)
                    Integ = (z1-self.z0)*(r1-self.r0)*(1-xi)+((z1-self.z0)*(r2-r1)+(z2-z1)*(r1-self.r0))*(1**2/2.-xi**2/2.)+(z2-z1)*(r2-r1)*(1**3/3.-xi**3/3.)
                    Mc=self.rho_w*self.g*self.lseg[s]*np.cross(Integ,n)
                else:
                    Mc=np.array([0.,0.,0.])
                M_ += Mc
        return M_

    # Returns a force acting on the iceberg from pressure drag 
    def Fdamp_qua(self,t,dot_r,dot_th):
        F_=np.array([0.,0.,0.])
        si=0
        for s in self.segments:
                si+=1
                Fc=np.array([0.,0.,0.])
                n,r1,r2=self.seg(s)
                z1=r1[1]
                z2=r2[1]
                eta = np.dot((dot_r + np.cross(dot_th,r1-self.r0)),n)
                beta = np.dot(np.cross(dot_th,r2-r1),n)
                if abs(beta)>1e-12:
                    xi_i = -eta/beta
                else:
                    xi_i = 2.e5
                if xi_i > 0 and xi_i < 1:  
                    if z1<=self.z0 and z2<=self.z0:
                        Fc=0.5*n*self.rho_w*self.lseg[s]*self.sign(eta) * ( (eta**2*xi_i + eta*beta*xi_i**2 + beta**2*xi_i**3/3.) - \
                                                                    (eta**2*(1-xi_i) + eta*beta*(1-xi_i**2) + beta**2*(1-xi_i**3)/3.))
                    elif z1<=self.z0 and z2>self.z0:
                        xi=(self.z0-z1)/(z2-z1)
                        if xi_i > xi:
                            Fc=0.5*n*self.rho_w*self.lseg[s]*self.sign(eta) * (eta**2*xi + eta*beta*xi**2 + beta**2*xi**3/3.)
                        else:
                            Fc=0.5*n*self.rho_w*self.lseg[s]*self.sign(eta) * ((eta**2*xi_i + eta*beta*xi_i**2 + beta**2*xi_i**3/3.) - \
                                                                       (eta**2*(xi-xi_i) + eta*beta*(xi**2-xi_i**2) + beta**2*(xi**3-xi_i**3)/3.)) 
                    elif z1>self.z0 and z2<=self.z0:
                        xi=(self.z0-z1)/(z2-z1)
                        if xi_i < xi:
                            Fc= - 0.5*n*self.rho_w*self.lseg[s]*self.sign(eta) * (eta**2*(1-xi) + eta*beta*(1-xi**2) + beta**2*(1-xi**3)/3.)
                        else:
                            Fc= - 0.5*n*self.rho_w*self.lseg[s]*self.sign(eta) * ( (eta**2*(1-xi_i) + eta*beta*(1-xi_i**2) + beta**2*(1-xi_i**3)/3.) - \
                                                                        (eta**2*(xi_i-xi) + eta*beta*(xi_i**2-xi**2) + beta**2*(xi_i**3-xi**3)/3.) )
                    else:
                        Fc=np.array([0.,0.,0.])
                else:
                    if z1<=self.z0 and z2<=self.z0:
                        Fc = 0.5*n*self.rho_w*self.lseg[s]*self.sign(eta) * (eta**2 + eta*beta + beta**2/3.)
                    elif z1<=self.z0 and z2>self.z0:
                        xi=(self.z0-z1)/(z2-z1)
                        Fc = 0.5*n*self.rho_w*self.lseg[s]*self.sign(eta) * (eta**2*xi + eta*beta*xi**2 + beta**2*xi**3/3.)
                    elif z1>self.z0 and z2<=self.z0:
                        xi=(self.z0-z1)/(z2-z1)
                        Fc = 0.5*n*self.rho_w*self.lseg[s]*self.sign(eta) * (eta**2*(1-xi) + eta*beta*(1-xi**2) + beta**2*(1-xi**3)/3.)
                    else:
                        Fc=np.array([0.,0.,0.])
                F_ -= Fc
        return self.alpha*F_  
    
    # Returns a moment acting on the iceberg from pressure drag and computed around its center of mass
    def Mdamp_qua(self,dot_r,dot_th):
        M_=np.array([0.,0.,0.])
        se=0
        for s in self.segments:
                Mc=np.array([0.,0.,0.])
                n,r1,r2=self.seg(s)
                z1=r1[1]
                z2=r2[1]
                eta = np.dot((dot_r + np.cross(dot_th,r1-self.r0)),n)
                beta = np.dot(np.cross(dot_th,r2-r1),n)
                if abs(beta)>1e-12:
                    xi_i = -eta/beta
                else:
                    xi_i = 2.e5
                v10 = np.cross(r1-self.r0,n)
                v21 = np.cross(r2-r1,n)
                C1 = eta**2*v10
                C2 = 2*eta*beta*v10 + eta**2*v21
                C3 = 2*eta*beta*v21 + beta**2*v10
                C4 = beta**2*v21
                if xi_i > 0 and xi_i < 1:     
                    if z1<=self.z0 and z2<=self.z0:
                        Mc=0.5*self.rho_w*self.lseg[s]*self.sign(eta) * ( \
                         C1*xi_i + C2*xi_i**2/2. + C3*xi_i**3/3. + C4*xi_i**4/4. -\
                        (C1*(1-xi_i) + C2*(1-xi_i**2)/2. + C3*(1-xi_i**3)/3. + C4*(1-xi_i**4)/4.) )
                        
                    elif z1<=self.z0 and z2>self.z0:
                        xi=(self.z0-z1)/(z2-z1)
                        if xi_i > xi:
                            Mc = 0.5*self.rho_w*self.lseg[s]*self.sign(eta) * ( C1*xi + C2*xi**2/2. + C3*xi**3/3. + C4*xi**4/4. )
                        else:
                            Mc = 0.5*self.rho_w*self.lseg[s]*self.sign(eta) * ( \
                              C1*xi_i + C2*xi_i**2/2. + C3*xi_i**3/3. + C4*xi_i**4/4. -\
                             (C1*(xi-xi_i) + C2*(xi**2-xi_i**2)/2. + C3*(xi**3-xi_i**3)/3. + C4*(xi**4-xi_i**4)/4.) )
                            
                    elif z1>self.z0 and z2<=self.z0:
                        xi=(self.z0-z1)/(z2-z1)
                        if xi_i < xi:
                            Mc = - 0.5*self.rho_w*self.lseg[s]*self.sign(eta) * (\
                            C1*(1-xi) + C2*(1-xi**2)/2. + C3*(1-xi**3)/3. + C4*(1-xi**4)/4. )
                                
                        else:
                            Mc = - 0.5*self.rho_w*self.lseg[s]*self.sign(eta) * (\
                                C1*(1-xi_i) + C2*(1-xi_i**2)/2. + C3*(1-xi_i**3)/3. + C4*(1-xi_i**4)/4. -\
                               (C1*(xi_i-xi) + C2*(xi_i**2-xi**2)/2. + C3*(xi_i**3-xi**3)/3. + C4*(xi_i**4-xi**4)/4.) )
                            
                    else:
                        Mc = np.array([0.,0.,0.])
                        
                else:
                    if z1<=self.z0 and z2<=self.z0: # done
                        Mc = 0.5*self.rho_w*self.lseg[s]*self.sign(eta) * (C1 + C2/2. + C3/3. + C4/4.)
                        
                    elif z1<=self.z0 and z2>self.z0:
                        xi=(self.z0-z1)/(z2-z1)
                        Mc = 0.5*self.rho_w*self.lseg[s]*self.sign(eta) * (C1*xi + C2*xi**2/2. + C3*xi**3/3. + C4*xi**4/4.)
                        
                    elif z1>self.z0 and z2<=self.z0:
                        xi=(self.z0-z1)/(z2-z1)
                        Mc = 0.5*self.rho_w*self.lseg[s]*self.sign(eta) * (C1*(1-xi) + C2*(1-xi**2)/2. + C3*(1-xi**3)/3. + C4*(1-xi**4)/4. )                    
                        
                    else:
                        Mc=np.array([0.,0.,0.])
                        
                M_ -= Mc
                se+=1
        return self.alpha*M_
    
#def Simu(eps, H, theta_0, rho_i, rho_w, dt, tmax, dz, alpha, cmx, cmz, cmI):  #theta in deg
def Simu(iceberg, dt, tmax, cmx, cmz, cmI):  #theta in deg
    
    Ntime=int(tmax/dt)
#    theta_0 *= np.pi/180. # Now, theta is in rad
#    # Compute initial vertical position of the center of mass of the iceberg
#    zg0 = -(iceberg.rho_i/iceberg.rho_w - 0.5)*iceberg.H*np.cos(iceberg.theta_0) + dz

    # Initialization. Call Iceberg's constructor.
    #iceberg = Iceberg( np.array([0.,zg0,0.]), [0,0,theta_0], eps*H, H, alpha, rho_i, rho_w)
    dot_r = np.array([0.,0.,0.])
    dot_theta = np.array([0.,0.,0.])
    rr = np.array([0.,0.,0.])
    F = np.array([0.,0.,0.])
    Ftotal = np.array([0.,0.,0.])
    F_ = np.array([0.,0.,0.])
    Mass = np.array(([0.,0.,0.],[0.,0.,0.],[0.,0.,0.]))
    Acc = np.array([0.,0.,0.])
    Madd = np.array(([0.,0.,0.],[0.,0.,0.],[0.,0.,0.]))
    
    # Initialize forces and moments
    F_ = iceberg.F() + iceberg.Fdamp_qua(0,dot_r,dot_theta)
    M_ = iceberg.M() + iceberg.Mdamp_qua(dot_r,dot_theta)



    # Initialize added mass 
    M11 = cmx*iceberg.Hmx()
    M22 = cmz*iceberg.Hmz()
    M66 = cmI*iceberg.Hinert()

    # Construct the full mass matrix
    M0 = np.array(([iceberg.mass,0.,0.],[0.,iceberg.mass,0.],[0.,0.,iceberg.inert]))
    Madd = np.array(([M11,0.,0.],[0.,M22,0.],[0.,0.,M66]))
    Mass = M0 + Madd
    if np.linalg.det(Mass)==0:
        print('A non inversible')
        return 0
    Mass_inv = np.linalg.inv(Mass)
    Fo = np.array(([F_[0]],[F_[1]],[M_[2]]))
    # Initial acceleration
    Acc = Mass_inv.dot(Fo)

    # Define output variables
    time = np.zeros(Ntime+1)
    xOut = np.zeros(Ntime+1) 
    zOut = np.zeros(Ntime+1)
    thetaOut = np.zeros(Ntime+1)
    Fx = np.zeros(Ntime+1)
    Fz = np.zeros(Ntime+1)
    My = np.zeros(Ntime+1)

    # Start integration in time
    it=0 # iteration = time step number
    for t in np.arange(0,tmax,dt):    
        # Compute current configuration: x,y,theta
        th = np.array([0,0,iceberg.theta[2] + (dot_theta[2] + dt*Acc[2]/2.)*dt])
        rr[0] = iceberg.r0[0] + (dot_r[0] + dt*Acc[0]/2.)*dt
        rr[1] = iceberg.r0[1] + (dot_r[1] + dt*Acc[1]/2.)*dt
        iceberg.theta = th.copy()    
        iceberg.r0 = rr.copy()

        # Evaluate forces and moments in the current configuration, select drag type
        F = iceberg.F() + iceberg.Fdamp_qua(t,dot_r,dot_theta)
        M = iceberg.M() + iceberg.Mdamp_qua(dot_r,dot_theta)

        # Adjust added masses and moment of inertia 
        M11 = cmx*iceberg.Hmx()
        M22 = cmz*iceberg.Hmz()
        M66 = cmI*iceberg.Hinert()
        # Construct the full mass matrix
        M0 = np.array(([iceberg.mass,0.,0.],[0.,iceberg.mass,0.],[0.,0.,iceberg.inert]))
        Madd = np.array(([M11,0.,0.],[0.,M22,0.],[0.,0.,M66]))
        Mass = M0 + Madd
        if np.linalg.det(Mass)==0:
            print('A non inversible')
            return 0
        Mass_inv = np.linalg.inv(Mass)
        Fi = 0.5*np.array([F[0]+F_[0],F[1]+F_[1],M[2]+M_[2]])
        Acc = Mass_inv.dot(Fi)
    
        # Compute velocities
        dot_r[0]  += dt*Acc[0]
        dot_r[1]  += dt*Acc[1]
        dot_theta += dt*np.array((0,0,Acc[2]))                  
        Ftotal = np.array([F[0],F[1],M[2]]) - Madd.dot(Acc)
   
        time[it] = t
        Fx[it] = Ftotal[0]
        Fz[it] = Ftotal[1]
        My[it] = Ftotal[2]
        xOut[it] = rr[0]
        zOut[it] = rr[1]
        thetaOut[it] = th[2]

        it += 1
        F_ = F
        M_ = M

    return time,xOut,zOut,thetaOut,Fx,Fz,My
